
package
{
	import flash.display.*;
	import flash.events.*;
	import flash.geom.*;
	
	import Box2D.Dynamics.*;
	import Box2D.Collision.*;
	import Box2D.Collision.Shapes.*;
	import Box2D.Common.Math.*;
	
	public class Alien extends Sprite
	{
		[Embed(source="images/alien4-big.png")]
		public static var bigAlienSrc: Class;
		
		[Embed(source="images/alien4.png")]
		public static var alien4Src: Class;
		
		public var body: b2Body;
		
		public var dead: Boolean = false;
		
		public var speed: Number = 3;
		
		public var type: uint;
		
		public function Alien (world: b2World, _x: Number, _y: Number, _type: uint = 5)
		{
			scaleX = 0.05;
			scaleY = 0.05;
			
			x = _x;
			y = _y;
			
			type = _type;
			
			var image: Bitmap;
			
			if (type == 5)
			{
				image = new bigAlienSrc();
				speed = 2;
			}
			else
			{
				image = new alien4Src();
				speed = 4;
			}
			
			image.x = -image.width / 2;
			image.y = -image.height / 2;
			
			addChild(image);
			
			// physics
			
			var bodyDef: b2BodyDef = new b2BodyDef();
			bodyDef.position.x = x / 32.0;
			bodyDef.position.y = y / 32.0;
			bodyDef.userData = this;
			bodyDef.fixedRotation = true;
			
			body = world.CreateBody(bodyDef);
			
			var circleDef: b2CircleDef = new b2CircleDef();
			circleDef.radius = image.width / 64.0;
			circleDef.density = 1.0;
			circleDef.friction = 0.0;
			circleDef.restitution = 0.2;
			
			body.CreateShape(circleDef);
			
			body.SetMassFromShapes();
		}
		
		public function update (player: Player, game: Game): void
		{
			var world: b2World = game.world;
			
			if (scaleX < ((type == 4) ? 1.5 : 1.0))
			{
				scaleX += 0.05;
				scaleY += 0.05;
			}
			
			var segment: b2Segment = new b2Segment();
			segment.p1 = body.GetPosition().Copy();
			segment.p2 = player.body.GetPosition().Copy();
			
			var bodies: Array = new Array();
			
			var count: int = world.Raycast(segment, bodies, 10, false, null);
			
			for (var i: uint = 0; i < count; i++)
			{
				var isStatic: Boolean = bodies[i].GetBody().IsStatic();
				
				if (isStatic) { return; }
			}
			
			if (Math.random() < 0.005 && type == 5 && ! game.player.dying)
			{
				var newAlien: Alien = new Alien(game.world, x, y, 4);
				
				game.addAlien(newAlien);
			}
			
			var target: * = player;
			
			var dx: Number = target.x - x;
			var dy: Number = target.y - y;
			
			var dz: Number = Math.sqrt(dx*dx + dy*dy);
			
			if (dz < speed)
			{
				// reached target
			}
			
			var velocity: b2Vec2 = new b2Vec2(dx * speed / dz, dy * speed / dz);
			
			body.SetLinearVelocity(velocity);
			
			rotation = Math.atan2(dy, dx) * 180 / Math.PI + 90;
		}
		
	}

}


