
package
{
	import flash.display.*;
	import flash.events.*;
	import flash.media.*;
	import flash.utils.*;
	
	import Box2D.Dynamics.*;
	import Box2D.Collision.*;
	import Box2D.Collision.Shapes.*;
	import Box2D.Common.Math.*;
	
	import mochi.as3.*;
	
	[SWF (width=640, height=480, frameRate=50)]
	public class Game extends Sprite
	{
		private var time: int;
		
		private var container: Sprite;
		private var container2: Sprite;
		
		private var aliensLayer: Sprite;
		private var particlesLayer: Sprite;
		
		private var particles: Array = new Array();
		private var bullets: Array = new Array();
		
		[Embed(source="images/bg.jpg")]
		public static var bgSrc: Class;
		
		[Embed(source="images/fg.jpg")]
		public static var fgSrc: Class;
		
		[Embed(source="images/title.png")]
		public static var titleSrc: Class;
		
		public var player: Player;
		
		public var aliens: Array = new Array();
		public var newAliens: Array = new Array();
		
		public var world: b2World;
		
		public var score: NumberTextField;
		public var scoreContainer: Sprite;
		
		public function Game ()
		{
			container = new Sprite();
			
			container2 = new Sprite();
			
			container2.x = 320;
			container2.y = 240;
			
			container2.addChild(container);
			
			addChild(container2);
			
			var titleImage: Bitmap = new titleSrc();
			
			titleImage.x = -titleImage.width / 2.0;
			titleImage.y = -titleImage.height / 2.0;
			
			container.addChild(titleImage);
			
			AudioControl.playMusic();
			
			createCave();
			
			world.SetContactListener(new MyContactListener());
			
			aliensLayer = new Sprite();
			container.addChild(aliensLayer);
			
			player = new Player(world);
			container.addChild(player);
			
			particlesLayer = new Sprite();
			container.addChild(particlesLayer);
			
			// set debug draw
			/*var dbgDraw:b2DebugDraw = new b2DebugDraw();
			var dbgSprite:Sprite = new Sprite();
			container.addChild(dbgSprite);
			dbgDraw.SetSprite(dbgSprite);
			dbgDraw.SetDrawScale(32);
			dbgDraw.SetFillAlpha(0.0);
			dbgDraw.SetLineThickness(1.0);
			dbgDraw.SetFlags(0xFFFFFFFF);
			world.SetDebugDraw(dbgDraw);*/
			
			score = new NumberTextField(0, 0);
			
			score.x = -score.width / 2.0;
			score.y = -score.height / 2.0;
			
			scoreContainer = new Sprite();
			
			scoreContainer.x = 320;
			scoreContainer.y = 40;
			
			scoreContainer.addChild(score);
			
			addChild(scoreContainer);
			
			addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		public function init (param : * = 0) : void
		{
			time = getTimer();
			
			stage.addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}
		
		private function onEnterFrame(e:Event):void
		{
			var dt:int = getTimer() - time;
			
			while (dt >= 20) {
				time += 20;
				dt -= 20;
				
				update();
			}
		}
		
		private function replay (param: * = null): void
		{
			stage.removeEventListener(Event.ENTER_FRAME, onEnterFrame);
			
			var rootObj: DisplayObjectContainer = parent;
			
			parent.addChild(new Game());
			
			parent.removeChild(this);
		}
		
		private function submit (param: * = null): void
		{
			var o:Object = { n: [4, 2, 14, 13, 7, 14, 3, 14, 4, 4, 1, 5, 8, 4, 2, 13], f: function (i:Number,s:String):String { if (s.length == 16) return s; return this.f(i+1,s + this.n[i].toString(16));}};
			var boardID:String = o.f(0,"");
			MochiScores.showLeaderboard({boardID: boardID, score: this.score.value});
		}
		
		public function update (): void
		{
			if (player.dying && ! player.dead)
			{
				player.dead = true;
				
				var gameOverText: MyTextField = new MyTextField(320, 100, "Game Over", "center", 60);
				
				//gameOverText.textColor = 0x1270e9;
				
				addChild(gameOverText);
				
				var replayButton: Button = new Button("Replay");
				
				replayButton.x = 320 - replayButton.width / 2.0;
				replayButton.y = 200;
				
				replayButton.addEventListener(MouseEvent.CLICK, replay);
				
				addChild(replayButton);
				
				var submitButton: Button = new Button("Submit score");
				
				submitButton.x = 320 - submitButton.width / 2.0;
				submitButton.y = 280;
				
				submitButton.addEventListener(MouseEvent.CLICK, submit);
				
				addChild(submitButton);
			}
			
			if (Math.random() < 0.02 && ! player.dying)
			{
				spawnAlien();
			}
			
			player.update(this);
			
			for (var blah: uint = 0; blah < newAliens.length; blah++)
			{
				aliens.push(newAliens[blah]);
			}
			
			newAliens = new Array();
			
			aliens = aliens.filter(updateAliensFilter);
			
			particles = particles.filter(updateParticlesFilter);
			
			world.Step(0.02, 10, 10);
			
			// Go through body list and update sprite positions/rotations
			for (var bb:b2Body = world.GetBodyList(); bb; bb = bb.GetNext()){
				if (bb.GetUserData() is DisplayObject){
					var obj:DisplayObject = bb.GetUserData();
					obj.x = bb.GetPosition().x * 32;
					obj.y = bb.GetPosition().y * 32;
					
					if (! bb.IsFixedRotation())
					{
						obj.rotation = bb.GetAngle() * (180/Math.PI);
					}
				}
			}
			
			bullets = bullets.filter(removeBulletsFilter);
			
			container.x = -player.x;
			container.y = -player.y;
			
			container2.rotation = -player.rotation;
			
			scoreContainer.rotation = player.rotation;
		}
		
		public function addParticle (p: Particle): void
		{
			particles.push(p);
			
			particlesLayer.addChild(p);
		}
		
		public function addBullet (b: Bullet): void
		{
			bullets.push(b);
			
			particlesLayer.addChild(b);
		}
		
		private function spawnAlien (): void
		{
			var ax: Number;
			var ay: Number;
			
			var i: uint;
			
			for (i = 0; i < 20; i++)
			{
				ax = player.x + Math.random() * 700 - 350;
				ay = player.y + Math.random() * 700 - 350;
				
				var dx: Number = player.x - ax;
				var dy: Number = player.y - ay;
				var dzSq: Number = dx*dx + dy*dy;
				
				if (dzSq < 150*150) 
				{
					continue;
				}
				
				var aabb:b2AABB = new b2AABB();
				aabb.lowerBound.Set((ax - 16) / 32.0, (ay - 16) / 32.0);
				aabb.upperBound.Set((ax + 16) / 32.0, (ay + 16) / 32.0);
				
				var k_bufferSize:Number = 10;
				var buffer:Array = [];
				var count:int = world.Query(aabb, buffer, k_bufferSize);
				
				var colliding: Boolean = false;
				
				for (var j: uint = 0; j < count; j++)
				{
					var body: b2Body = buffer[j].GetBody();
					
					if (! body.IsStatic())
					{
						colliding = true;
						break; // inner loop
					}
					
					var vertexCount: int = buffer[j].GetVertexCount();
					var vertices: Array = buffer[j].GetVertices();
					
					if (circleTouchingPolygon(new b2Vec2(ax / 32.0, ay / 32.0), 0.5, vertices, vertexCount))
					{
						colliding = true;
						break; // inner loop
					}
				}
				
				if (! colliding)
				{
					break;
				}
			}
			
			if (i >= 20)
			{
				return;
			}
			
			var alien: Alien = new Alien(world, ax, ay);
			
			addAlien(alien);
		}
		
		public function addAlien (alien: Alien): void
		{
			newAliens.push(alien);
			
			aliensLayer.addChild(alien);
		}
		
		private static function circleTouchingPolygon (c: b2Vec2, r: Number, vertices: Array, count: int): Boolean
		{
			for (var i: uint = 0; i < count; i++)
			{
				var j: uint = (i + 1) % count;
				
				var normal: b2Vec2 = new b2Vec2(vertices[j].y - vertices[i].y, vertices[i].x - vertices[j].x);
				
				var nDotC: Number = c.x * normal.x + c.y * normal.y;
				
				var nDotI: Number = vertices[i].x * normal.x + vertices[i].y * normal.y;
				
				var overlap: Number = nDotI - nDotC + r;
				
				if (overlap < 0)
				{
					return false;
				}
			}
			
			return true;
		}
		
		private function updateParticlesFilter (p: Particle, index: int, arr: Array): Boolean
		{
			p.update();
			
			if (p.dead)
			{
				// Remove from display list
				particlesLayer.removeChild(p);
				
				// And remove from array
				return false;
			}
			else
			{
				// p remains in array
				return true;
			}
		}
		
		private function removeBulletsFilter (b: Bullet, index: int, arr: Array): Boolean
		{
			if (b.dead)
			{
				world.DestroyBody(b.body);
				
				// Remove from display list
				particlesLayer.removeChild(b);
				
				// And remove from array
				return false;
			}
			else
			{
				// p remains in array
				return true;
			}
		}
		
		private function updateAliensFilter (alien: Alien, index: int, arr: Array): Boolean
		{
			alien.update(player, this);
			
			if (alien.dead)
			{
				world.DestroyBody(alien.body);
				
				AudioControl.playAlienDeathSound();
				
				score.value += (alien.type == 4) ? 50 : 100;
				
				// Remove from display list
				aliensLayer.removeChild(alien);
				
				for (var i: uint = 0; i < 30; i++)
				{
					addParticle(new AlienParticle(alien));
				}
				
				// And remove from array
				return false;
			}
			else
			{
				// alien remains in array
				return true;
			}
		}
		
		private function digCave (vertices: Array, oldCentre: b2Vec2, olderCentres: Array, worldAABB:b2AABB): void
		{
			var start: b2Vec2 = vertices[vertices.length - 1];
			
			var centre: b2Vec2 = new b2Vec2(2 * start.x - oldCentre.x, 2 * start.y - oldCentre.y);
			
			for (var i: uint = 0; i < olderCentres.length; i++)
			{
				if (olderCentres[i] == oldCentre)
				{
					continue;
				}
				
				var dx: Number = olderCentres[i].x - centre.x;
				var dy: Number = olderCentres[i].y - centre.y;
				var dzSq: Number = dx*dx + dy*dy;
				
				if (dzSq < 30*30)
				{
					return;
				}
			}
			
			olderCentres.push(centre);
			
			var startAngle: Number = Math.atan2(start.y - centre.y, start.x - centre.x);
			
			for (i = 0; i < 8; i++)
			{
				var angle: Number = i * Math.PI / 4.0 + startAngle;
				
				angle += Math.random() * Math.PI / 4.0 - Math.PI / 8.0;
				
				var dist: Number = Math.random() * 3 + 6;
				
				var _x: Number = centre.x + Math.cos(angle) * dist + Math.random() * 2 - 1;
				var _y: Number = centre.y + Math.sin(angle) * dist + Math.random() * 2 - 1;
				
				vertices.push(new b2Vec2(_x, _y));
				
				worldAABB.lowerBound.x = Math.min(worldAABB.lowerBound.x, _x - 10);
				worldAABB.lowerBound.y = Math.min(worldAABB.lowerBound.y, _y - 10);
				
				worldAABB.upperBound.x = Math.max(worldAABB.upperBound.x, _x + 10);
				worldAABB.upperBound.y = Math.max(worldAABB.upperBound.y, _y + 10);
				
				//if (Math.random() < 0.3)
				{
					digCave(vertices, centre, olderCentres, worldAABB);
				}
			}
		}
		
		private function createCave (): void
		{
			// Creat world AABB
			var worldAABB:b2AABB = new b2AABB();
			worldAABB.lowerBound.Set(-30.0, -30.0);
			worldAABB.upperBound.Set(30.0, 30.0);
			
			var vertices: Array = new Array();
			var centres: Array = new Array();
			centres.push(new b2Vec2(0, 0));
			
			for (var i: uint = 0; i < 8; i++)
			{
				var angle: Number = i * Math.PI / 4.0;
				
				angle += Math.random() * Math.PI / 4.0 - Math.PI / 8.0;
				
				var dist: Number = Math.random() * 6 + 10;
				
				var _x: Number = Math.cos(angle) * dist + Math.random() * 2 - 1;
				var _y: Number = Math.sin(angle) * dist + Math.random() * 2 - 1;
				
				vertices.push(new b2Vec2(_x, _y));
				
				//if (i % 2 == 0)
				{
					digCave(vertices, centres[0], centres, worldAABB);
				}
			}
			
			var normals: Array = new Array();
			
			for (i = 0; i < vertices.length; i++)
			{
				var j: uint = (i + 1) % vertices.length;
	    		
	    		var normalIJ: b2Vec2 = new b2Vec2(vertices[j].y - vertices[i].y, vertices[i].x - vertices[j].x);
	    		normalIJ.Normalize();
	    		
	    		normals.push(normalIJ);
			}
			
			// Define the gravity vector
			var gravity:b2Vec2 = new b2Vec2(0.0, 0.0);
			
			// Allow bodies to sleep
			var doSleep:Boolean = false;
			
			// Construct a world object
			world = new b2World(worldAABB, gravity, doSleep);
			
			// Add ground body
			var bodyDef: b2BodyDef = new b2BodyDef();
			var polyDef: b2PolygonDef = new b2PolygonDef();
			polyDef.friction = 0.0;
			polyDef.density = 0; // static bodies require zero density
			
			for (i = 0; i < vertices.length; i++)
	    	{
	    		//bodyDef.position.Set(-12, 0);
	    		
	    		j = (i + 1) % vertices.length;
	    		var k: uint = (i + 2) % vertices.length;
	    		var l: uint = (i + 3) % vertices.length;
	    		
	    		normalIJ = normals[i];
	    		var normalJK: b2Vec2 = normals[j];
	    		var normalKL: b2Vec2 = normals[k];
	    		
	    		var myNewVertices: Array = new Array();
				
				myNewVertices[0] = new b2Vec2(vertices[k].x, vertices[k].y);
				myNewVertices[1] = new b2Vec2(vertices[j].x, vertices[j].y);
				myNewVertices[2] = new b2Vec2(vertices[j].x + normalIJ.x * 9 + normalJK.x * 9, vertices[j].y + normalIJ.y * 9 + normalJK.y * 9);
				myNewVertices[3] = new b2Vec2(vertices[k].x + normalJK.x * 9 + normalKL.x * 9, vertices[k].y + normalJK.y * 9 + normalKL.y * 9);
	    		
				polyDef.vertices = convexHull(myNewVertices);
				polyDef.vertexCount = polyDef.vertices.length;
				
				var body: b2Body = world.CreateBody(bodyDef);
				body.CreateShape(polyDef);
				body.SetMassFromShapes();
	    	}
			
			// draw cave
			
			var minX: Number = worldAABB.lowerBound.x;
			var minY: Number = worldAABB.lowerBound.y;
			var caveW: Number = worldAABB.upperBound.x - minX;
			var caveH: Number = worldAABB.upperBound.y - minY;
			
			container.graphics.beginBitmapFill(new bgSrc().bitmapData);
			container.graphics.drawRect(minX * 32, minY * 32, caveW * 32, caveH * 32);
			container.graphics.endFill();
			
			//container.graphics.beginBitmapFill(new fgSrc().bitmapData);
			for (var b:b2Body = world.GetBodyList(); b; b = b.GetNext())
			{
			    for (var s:b2Shape = b.GetShapeList(); s; s = s.GetNext())
				{
				    if (s is b2PolygonShape)
				    {
				    	var p: b2PolygonShape = s as b2PolygonShape;
				    	
				    	var polygonVertices: Array = p.GetVertices();
				    	
				    	container.graphics.beginBitmapFill(new fgSrc().bitmapData);
				    	//container.graphics.beginFill(0xFFFFFF, 0.1);
				    	
				    	var vertex: b2Vec2 = b.GetWorldPoint(polygonVertices[0]);
				    	container.graphics.moveTo(vertex.x * 32, vertex.y * 32);
				    	
				    	for (i = 1; i < p.GetVertexCount(); i++)
				    	{
				    		vertex = b.GetWorldPoint(polygonVertices[i]);
				    		container.graphics.lineTo(vertex.x * 32, vertex.y * 32);
				    	}
				    	
				    	container.graphics.endFill();
				    	
				    	// outline
				    	
				    	/*container.graphics.lineStyle(2, 0xb55217);
				    	
				    	vertex = b.GetWorldPoint(polygonVertices[p.GetVertexCount() - 1]);
				    	container.graphics.moveTo(vertex.x * 32, vertex.y * 32);
				    	
				    	for (i = 0; i < p.GetVertexCount(); i++)
				    	{
				    		vertex = b.GetWorldPoint(polygonVertices[i]);
				    		container.graphics.lineTo(vertex.x * 32, vertex.y * 32);
				    	}*/
				    }
				}
			}
			//container.graphics.endFill();
			
			/*var colours: Array = new Array(0xb55217, 0x903d18, 0x471e0c);
			var alphas: Array = new Array(1, 1, 1);
			var ratios: Array = new Array(0, 0, 255);
			
			container.graphics.lineStyle(2, 0xb55217);
			//container.graphics.lineGradientStyle(GradientType.LINEAR, colours, alphas, ratios, null, SpreadMethod.REFLECT);
			
			vertex = vertices[0];
	    	container.graphics.moveTo(vertex.x * 32, vertex.y * 32);
			
			for (i = 1; i < vertices.length; i++)
	    	{
	    		vertex = vertices[i];
	    		container.graphics.lineTo(vertex.x * 32, vertex.y * 32);
	    	}*/
		}
		
		// stolen from http://www.bit-101.com/blog/?p=1497
		// I don't even need this, it just saves me thinking
		private function convexHull(points:Array):Array
		{
			points.sortOn("x", Array.NUMERIC);
			var point:b2Vec2 = points[0] as b2Vec2;
			var hull:Array = new Array();
			var angle:Number = Math.PI / 2;
			while (point != hull[0])
			{
				hull.push(point);
				var bestAngle:Number = Number.MAX_VALUE;
				var bestIndex:int;
				for (var i:int = 0; i <points.length; i++)
				{
					var testPoint:b2Vec2 = points[i] as b2Vec2;
					if (testPoint == point)
					{
						continue;
					}
					var dx:Number = testPoint.x - point.x;
					var dy:Number = testPoint.y - point.y;
					var testAngle:Number = Math.atan2(dy,dx);
					testAngle -= angle;
					while (testAngle <0)
					{
						testAngle+=Math.PI*2;
					}
					if (testAngle<bestAngle)
					{
						bestAngle=testAngle;
						bestIndex=i;
					}
				}
				point=points[bestIndex];
				angle+=bestAngle;
			}
			return hull;
		}
	}

}


