
package
{
	import flash.display.*;
	import flash.events.*;
	import flash.geom.*;
	
	import Box2D.Dynamics.*;
	import Box2D.Collision.*;
	import Box2D.Collision.Shapes.*;
	import Box2D.Common.Math.*;
	
	public class Player extends Sprite
	{
		[Embed(source="images/player.png")]
		public static var playerSrc: Class;
		
		public var body: b2Body;
		
		private var arm: Shape;
		
		private var armOffset: Number = 1;
		private var armLength: Number = 8;
		
		private var blow: Boolean = false;
		private var shootDown: Boolean = false;
		
		private var timeToShoot: uint = 0;
		
		public var dead: Boolean = false;
		
		public var dying: b2Vec2 = null;
		
		public static const A: uint = 65;
		public static const C: uint = 67;
		public static const S: uint = 83;
		public static const X: uint = 88;
		public static const Y: uint = 89;
		public static const Z: uint = 90;
		
		public function Player (world: b2World)
		{
			var image: Bitmap = new playerSrc()
			
			image.x = -image.width / 2;
			image.y = -image.height / 2;
			
			addChild(image);
			
			arm = new Shape();
			
			arm.y = -armOffset;
			
			arm.graphics.beginFill(0x8b8b8b);
			arm.graphics.drawRect(-1, -1, armLength, 2);
			arm.graphics.endFill();
			
			addChild(arm);
			
			/*graphics.lineStyle(1, 0xFFFFFF);
			graphics.drawCircle(0, 0, 150);*/
			
			// physics
			
			var bodyDef: b2BodyDef = new b2BodyDef();
			bodyDef.position.x = 0;
			bodyDef.position.y = 0;
			bodyDef.userData = this;
			bodyDef.angularDamping = 0.1;
			
			body = world.CreateBody(bodyDef);
			
			var circleDef: b2CircleDef;
			
			// head
			
			/*var circleDef: b2CircleDef = new b2CircleDef();
			circleDef.radius = 4.0 / 32.0;
			circleDef.density = 1.0;
			circleDef.friction = 0.0;
			circleDef.restitution = 0.2;
			circleDef.localPosition = new b2Vec2(0, -8.0 / 32.0);
			
			body.CreateShape(circleDef);
			
			// body
			
			var boxDef: b2PolygonDef = new b2PolygonDef();
			boxDef.vertexCount = 4;
			boxDef.vertices[0].Set(-3.0 / 32.0, -4.0 / 32.0);
			boxDef.vertices[1].Set( 3.0 / 32.0, -4.0 / 32.0);
			boxDef.vertices[2].Set( 3.0 / 32.0,  7.0 / 32.0);
			boxDef.vertices[3].Set(-3.0 / 32.0,  7.0 / 32.0);
			boxDef.density = 1.0;
			boxDef.friction = 0.0;
			boxDef.restitution = 0.2;
			
			body.CreateShape(boxDef);
			
			// backpack
			
			boxDef.vertexCount = 4;
			boxDef.vertices[0].Set( 5.0 / 32.0, -5.0 / 32.0);
			boxDef.vertices[1].Set(10.0 / 32.0, -5.0 / 32.0);
			boxDef.vertices[2].Set(10.0 / 32.0,  6.0 / 32.0);
			boxDef.vertices[3].Set( 5.0 / 32.0,  6.0 / 32.0);
			boxDef.density = 2.0;
			
			body.CreateShape(boxDef);*/
			
			// actual collision shape
			
			circleDef = new b2CircleDef();
			circleDef.radius = 0.5;
			circleDef.density = 1.0;
			circleDef.friction = 0.0;
			circleDef.restitution = 0.2;
			circleDef.localPosition = new b2Vec2(0, 0);
			circleDef.filter.groupIndex = -1; // never collide with bullets
			
			var circle: b2Shape = body.CreateShape(circleDef);
			
			body.SetMassFromShapes();
			
			body.DestroyShape(circle);
			
			circleDef.radius = 24.0 / 64.0;
			
			body.CreateShape(circleDef);
			
			addEventListener(Event.ADDED_TO_STAGE, init);
			addEventListener(Event.REMOVED_FROM_STAGE, uninit);
		}
		
		public function update (game: Game): void
		{
			if (dying)
			{
				doBleed(game);
				
				return;
			}
			
			/*var angularVelocity: Number = body.GetAngularVelocity();
			
			body.ApplyTorque(-angularVelocity * 0.02);*/
			
			var angle: Number = Math.atan2(mouseY, mouseX);
			
			arm.rotation = angle * 180 / Math.PI;
			
			if (blow)
			{
				doBlow(game);
			}
			
			if (timeToShoot == 0)
			{
				if (shootDown)
				{
					doRecoil(2, game);
					timeToShoot = 10;
				}
			}
			else
			{
				timeToShoot -= 1;
			}
		}
		
		private function doRecoil (recoil: Number, game: Game): void
		{
			var mouseLocal: b2Vec2 = new b2Vec2(mouseX, mouseY + armOffset);
			
			var mouseWorld: b2Vec2 = body.GetWorldVector(mouseLocal);
			
			mouseWorld.Normalize();
			
			var force: b2Vec2 = mouseWorld.Negative();
			
			force.Multiply(recoil);
			
			body.ApplyForce(force, body.GetPosition());
			
			var emitter: b2Vec2 = body.GetWorldPoint(new b2Vec2(0, -armOffset / 32.0));
			emitter.x += mouseWorld.x * armLength / 32.0;
			emitter.y += mouseWorld.y * armLength / 32.0;
			
			force.Multiply(0.4);
			
			var com: b2Vec2 = body.GetWorldPoint(new b2Vec2(6.0 / 32.0, 0)); // centre of mass
			
			var torque: Number = ((emitter.x - com.x) * force.y - (emitter.y - com.y) * force.x);
			
			body.ApplyTorque(torque);
			
			var velocity: b2Vec2 = body.GetLinearVelocityFromWorldPoint(emitter);
			velocity.x += mouseWorld.x * 10;
			velocity.y += mouseWorld.y * 10;
			
			var bullet: Bullet = new Bullet(game.world, emitter, velocity);
			
			game.addBullet(bullet);
		}
		
		private function doBleed (game: Game): void
		{
			var emitter: b2Vec2 = body.GetWorldPoint(dying);
			
			/*var force: b2Vec2 = body.GetWorldVector(new b2Vec2(-0.1, 0));
			
			var com: b2Vec2 = body.GetWorldPoint(new b2Vec2(6.0 / 32.0, 0)); // centre of mass
			
			var torque: Number = ((emitter.x - com.x) * force.y - (emitter.y - com.y) * force.x);
			
			body.ApplyTorque(torque);*/
			
			var velocity: b2Vec2 = body.GetLinearVelocityFromWorldPoint(emitter);
			velocity.Multiply(32.0 / 50.0);
			
			emitter.Multiply(32.0);
			
			var p: Particle;
			
			p = new BloodParticle(emitter.x, emitter.y, velocity.x, velocity.y);
			game.addParticle(p);
			
			/*p = new BloodParticle(emitter.x, emitter.y, velocity.x, velocity.y);
			game.addParticle(p);
			
			p = new BloodParticle(emitter.x, emitter.y, velocity.x, velocity.y);
			game.addParticle(p);*/
			
			dying.x *= 0.95;
			dying.y *= 0.95;
		}
		
		private function doBlow (game: Game): void
		{
			var mouseLocal: b2Vec2 = new b2Vec2(mouseX, mouseY + armOffset);
			
			var mouseWorld: b2Vec2 = body.GetWorldVector(mouseLocal);
			
			mouseWorld.Normalize();
			
			var force: b2Vec2 = mouseWorld.Negative();
			
			force.Multiply(2);
			
			body.ApplyForce(force, body.GetPosition());
			
			var emitter: b2Vec2 = body.GetWorldPoint(new b2Vec2(0, -armOffset / 32.0));
			emitter.x += mouseWorld.x * armLength / 32.0;
			emitter.y += mouseWorld.y * armLength / 32.0;
			
			var com: b2Vec2 = body.GetWorldPoint(new b2Vec2(6.0 / 32.0, 0)); // centre of mass
			
			var torque: Number = ((emitter.x - com.x) * force.y - (emitter.y - com.y) * force.x);
			
			body.ApplyTorque(torque);
			
			var velocity: b2Vec2 = body.GetLinearVelocityFromWorldPoint(emitter);
			velocity.Multiply(32.0 / 50.0);
			velocity.Add(mouseWorld);
			
			emitter.Multiply(32.0);
			
			var p: Particle;
			
			p = new AirParticle(emitter.x, emitter.y, velocity.x, velocity.y);
			game.addParticle(p);
			
			p = new AirParticle(emitter.x, emitter.y, velocity.x, velocity.y);
			game.addParticle(p);
			
			p = new AirParticle(emitter.x, emitter.y, velocity.x, velocity.y);
			game.addParticle(p);
		}
		
		private function init (param : * = null): void
		{
			stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDown);
			stage.addEventListener(KeyboardEvent.KEY_UP, onKeyUp);
		}
		
		private function uninit (param : * = null): void
		{
			stage.removeEventListener(KeyboardEvent.KEY_DOWN, onKeyDown);
			stage.removeEventListener(KeyboardEvent.KEY_UP, onKeyUp);
		}
		
		private function onKeyDown (e: KeyboardEvent): void
		{
			if (e.keyCode == Z || e.keyCode == Y || e.keyCode == A || e.keyCode == C)
			{
				blow = true;
			}
			else if (e.keyCode == X || e.keyCode == S)
			{
				shootDown = true;
			}
		}
		
		private function onKeyUp (e: KeyboardEvent): void
		{
			if (e.keyCode == Z || e.keyCode == Y || e.keyCode == A || e.keyCode == C)
			{
				blow = false;
			}
			else if (e.keyCode == X || e.keyCode == S)
			{
				shootDown = false;
			}
		}
	}

}


